import java.util.Arrays;
import java.io.*;

public class CpmCreateFile {
	private int user;
	private String filename;

	private byte[] dir;
	private long dirOff;
	private RandomAccessFile fd = null;
	private CpmDpb dpb;
	private InputStream data;
	private int[] sectrn = null;

	private int nextDirEnt; // index, not disk offset
	private int nextBlock; // index, not disk offset
	private int maxDirEnt;
	private int maxBlock;
	private boolean smallAlloc;
	private int blkSize;
	private int blkRecs;
	private int curRC;
	private int curExt;
	private int curS2;
	private int curEnt;
	private int curDirEnt;
	private int curDirOff;
	private int curBlock;

	public CpmCreateFile(File disk) {
		try {
			fd = new RandomAccessFile(disk, "rw");
		} catch (Exception ee) {
			ee.printStackTrace();
			return;
		}
	}

	public boolean setDpb(CpmDpb dpb) {
		dir = new byte[(dpb.drm + 1) * 32];
		dirOff = (dpb.off * dpb.spt + dpb.poff) * 128; // bytes
		try {
			fd.seek(dirOff);
			int n = fd.read(dir);
			if (n < 0) {
				System.err.format("EOF reading directory\n");
				return false;
			}
		} catch (Exception ee) {
			ee.printStackTrace();
			return false;
		}
		// Disk must be blank, initially.
		// It should have been just formatted, so not a problem.
		for (int x = 0; x <= dpb.drm; ++x) {
			if ((dir[x * 32] & 0xff) != 0xe5) {
				System.err.format("Directory not empty (%d, %d, %d)\n", x, dirOff, dir.length);
				return false;
			}
		}
		int b = Integer.numberOfTrailingZeros(dpb.al0);
		if (b > 15) {
			System.err.format("DPB ALV0 not valid\n");
			return false;
		}
		b = 16 - b;
		this.dpb = dpb;
		nextDirEnt = 0;
		maxDirEnt = dpb.drm;
		nextBlock = b;
		maxBlock = dpb.dsm;
		smallAlloc = (dpb.dsm < 256);
		blkSize = (128 << dpb.bsh);
		blkRecs = dpb.blm + 1;
		if (dpb.plat > 1) {
			buildSecTab();
		}
		return true;
	}

	private void buildSecTab() {
		// operate translation in CP/M records...
		int pspl = (dpb.pssz / 128);
		sectrn = new int[dpb.spt];
		boolean[] secflg = new boolean[dpb.spt];
		Arrays.fill(sectrn, -1);
		Arrays.fill(secflg, false);
		int ntran = 0;
		int s = 0;
		while (ntran < dpb.spt) {
			while (secflg[s]) {
				s += pspl;
				if (s >= dpb.spt) s -= dpb.spt;
			}
			for (int x = 0; x < pspl; ++x) {
				secflg[s] = true;
				sectrn[ntran++] = s++;
			}
			s += (dpb.plat - 1) * pspl;
			if (s >= dpb.spt) s -= dpb.spt;
		}
//		String str = "sectrn[] = {";
//		for (int x = 0; x < dpb.spt; ++x) {
//			str += String.format(" %d", sectrn[x]);
//		}
//		System.err.format("%s }\n", str);
	}

	public boolean isError() { return fd == null; }

	public boolean putFile(File fi) {
		data = null;
		if (!fi.exists()) {
			return false;
		}
		try {
			data = new FileInputStream(fi);
		} catch (Exception ee) {
			return false;
		}
		curRC = 0;
		curExt = 0;
		curS2 = 0;
		setupFile(fi.getName());
		boolean ret = true;
		if (!makeFileEnt() || !writeFile(data)) {
			ret = false;
		}
		closeFile();
		return ret;
	}

	public boolean close() {
		if (fd != null) {
			if (sectrn == null) {
				try {
					fd.seek(dirOff);
					fd.write(dir);
					fd.close();
				} catch (Exception ee) {
					return false;
				}
			} else {
				int dblks = (((dpb.drm + 1) / 4) + dpb.blm) >> dpb.bsh;
				int off = 0;
				for (int blk = 0; blk < dblks; ++blk) {
					writeBlock(blk, blkSize, dir, off);
					off += (128 << dpb.bsh);
				}
			}
		}
		return true;
	}

	private void setupFile(String name) {
		byte[] nam = new byte[11];
		int t = name.indexOf(':');
		if (t > 0) {
			try {
				user = Integer.valueOf(name.substring(0, t));
			} catch(Exception ee) {
				user = 0;
			}
			++t;
		} else {
			user = 0;
			t = 0;
		}
		int x = 0;
		while (t < name.length() && name.charAt(t) != '.' && x < 8) {
			nam[x++] = (byte)Character.toUpperCase(name.charAt(t++));
		}
		while (x < 8) {
			nam[x++] = ' ';
		}
		while (t < name.length() && name.charAt(t) != '.') {
			++t;
		}
		if (t < name.length() && name.charAt(t) == '.') {
			++t;
		}
		while (t < name.length() && name.charAt(t) != '.' && x < 11) {
			nam[x++] = (byte)Character.toUpperCase(name.charAt(t++));
		}
		while (x < 11) {
			nam[x++] = ' ';
		}
		filename = new String(nam);
	}

	private void setBlock(int blk, int bytes) {
		int nrecs = bytes / 128;
		int doff = curDirOff + 12; // point to EXT
		// update current values in dirent before proceding...
		curRC += nrecs;
		dir[doff++] = (byte)curExt;
		++doff;
		dir[doff++] = (byte)curS2;
		dir[doff++] = (byte)curRC;
		// doff now points to DM[0]...
		doff += curEnt;
		++curEnt;
		dir[doff++] = (byte)blk;
		if (!smallAlloc) {
			dir[doff++] = (byte)(blk >> 8);
			++curEnt;
		}
		// now prepare for (possible) next write...
		if (curRC >= 128) {
			// "open reel"...
			curRC = 0;
			curExt = (curExt + 1) & 0x1f;
			if (curExt != 0) {
				if ((curExt & dpb.exm) == 0) {
					curEnt = 16; // just in case
				}
			} else {
				curEnt = 16; // just in case
				// starting new "module"...
				curS2 += 1;
				if ((curS2 & 0x0f) == 0) {
					System.err.format("File too large\n");
					// what next?
					curRC = 128; // signal error
				}
			}
		}
	}

	private void closeFile() {
		if (data != null) {
			try {
				data.close();
			} catch (Exception ee) {}
			data = null;
		}
	}

	private boolean makeFileEnt() {
		if (nextDirEnt > maxDirEnt) {
			System.err.format("Directory full\n");
			return false;
		}
		curDirEnt = nextDirEnt++;
		curDirOff = curDirEnt * 32;
		int doff = curDirOff;
		dir[doff++] = (byte)(user & 0x1f);
		byte[] fn = filename.getBytes();
		System.arraycopy(fn, 0, dir, doff, 11);
		doff += 11; // start+12, EXT
		Arrays.fill(dir, doff, doff + 20, (byte)0);
		// NOTE: "dirty" bit in S2 is for user only, not dir entry on disk.
		//doff += 2; // start+14, S2
		//dir[doff] |= 0x80;
		dir[doff++] = (byte)curExt;
		++doff;
		dir[doff++] = (byte)curS2;
		curEnt = 0;
		return true;
	}

	// Must handle possible logical-to-physical sector translation...
	private boolean writeBlock(int blk, int num, byte[] buf, int boff) {
		if (sectrn == null) {
			long off = ((blk << dpb.bsh) * 128) + dirOff;
			try {
				fd.seek(off);
				fd.write(buf, boff, num);
			} catch (Exception ee) {
				return false;
			}
			return true;
		}
		// we know block 0 starts a track, but blocks may span 2 tracks...
		int recoff = (blk << dpb.bsh);
		for (int r = 0; r <= dpb.blm && num > 0; ++r) {
			int ps = (recoff + r) % dpb.spt;
			int toff = ((recoff + r) / dpb.spt) * dpb.spt;
			int ls = sectrn[ps];
			long off =  (toff + ls) * 128 + dirOff;
			try {
				fd.seek(off);
				fd.write(buf, boff, 128);
				boff += 128;
				num -= 128;
			} catch (Exception ee) {
				return false;
			}
		}
		return true;
	}

	private boolean writeFile(InputStream data) {
		byte[] buf = new byte[blkSize];
		while (true) {
			int w = blkSize;
			int n;
			try {
				n = data.read(buf);
				if (n < 0) {
					break;
				}
			} catch (Exception ee) {
				return false;
			}
			if (n < blkSize) {
				w = (n + 127) / 128;
				w *= 128;
				if (n < w) {
					Arrays.fill(buf, n, w, (byte)0x1a);
				}
			}
			if (nextBlock > maxBlock) {
				System.err.format("Disk full\n");
				return false;
			}
			if (curEnt >= 16) {
				if (!makeFileEnt()) {
					return false;
				}
				curEnt = 0;
			}
			curBlock = nextBlock++;
			if (!writeBlock(curBlock, w, buf, 0)) {
				return false;
			}
			setBlock(curBlock, w);
		}
		return true;
	}

	public void doPreload(File pl, CpmDpb dpb) {
		if (!setDpb(dpb)) {
			System.err.format("Bad DPB, Skipping %s\n", pl.getName());
			return;
		}
		File[] ents = pl.listFiles();
		int n = 0;
		for (File f : ents) {
			if (f.isDirectory()) {
				continue;
			}
			if (!putFile(f)) {
				System.err.format("Failed to add file %s to image\n",
									f.getName());
				continue;
			}
			++n;
		}
		if (n == 0) {
			System.err.format("No files loaded %s\n",
					pl.getAbsolutePath());
		} else {
			System.out.format("Preloaded partition from %s, %d files\n",
					pl.getAbsolutePath(), n);
		}
	}
}
