import java.util.Arrays;
import java.io.*;

public class fmtsasi {

	private void usage() {
		System.err.format("Usage: fmtsasi [options] <image> [format-options]\n"+
		                  "       fmtsasi list\n"+
		                  "       fmtsasi dump <image>\n"+
			"Options:\n" +
			"    list    = List all drive types and capacities, then exit\n" +
			"    dump    = Show format of existing image\n" +
			"    bsh=n   = set minimum BSH for CP/M DPBs\n" +
			"    drm=f   = factor of dir entries to min, >= 1.0\n" +
			"    lat=n   = describe sector latency layout (unused)\n" +
			"    par=size = add partition size <size> sectors\n" +
			"              Last may be 0 for reset of disk\n" +
			"    files=[p,]d = preload image (partition 'p') " +
					"with files from directory 'd'\n" +
			"    force   = overwrite an existing image\n" +
			"    nopart  = Do not create MMS partition table\n" +
			"    repair  = overwrite only partition info of existing image\n" +
			"format-Options:\n" +
			"    XEBEC_ST506 |\n" +
			"        XEBEC_ST412 |\n" +
			"        XEBEC_CM5206 |\n" +
			"        XEBEC_CM5410 |\n" +
			"        XEBEC_CM5616 |\n" +
			"        XEBEC_RO201 |\n" +
			"        XEBEC_RO202 |\n" +
			"        XEBEC_RO203 |\n" +
			"        XEBEC_RO204\n" +
			"    SS\n" +
			"Where: SS = Sector Size 512 (default 256)\n" +
			"       Partition default is a single partition full disk\n"
			);
		System.exit(1);
	}

	private void dump(File fi) {
		if (!fi.exists()) {
			System.err.format("No file: %s\n", fi.getAbsolutePath());
			System.exit(1);
		}
		H89SASIFormatter ff = new H89SASIFormatter(fi);
		if (ff.isError()) {
			System.err.format("Not a valid image file: %s\n", fi.getAbsolutePath());
			System.exit(1);
		}
		ff.dump(System.out);
		System.exit(0);
	}

	private void list() {
		H89SASIFormatter ff;
		for (int dsk = 1; true; ++dsk) {
			ff = new H89SASIFormatter(dsk, 512, 1, 3, 1.0f);
			if (ff.isError()) {
				break;
			}
			double cap = ff.getCapacity(); // bytes...
			cap /= 1024;
			cap /= 1024;
			System.out.format("%14s %6.1f MB\n", H89SASIFormatter.disks[dsk], cap);
		}
		System.exit(0);
	}

	// 'start' and 'psize' are in disk sectors (ssz), not CP/M records (128).
	private CpmDpb setcpmdpb(H89SASIFormatter ff,
				int start, int psize, int ssz, int minBsh, float factor) {
		CpmDpb dpb = new CpmDpb();
		int cpmsecs = (psize * ssz) / 128;
		cpmsecs -= ff.resv_trks * 64; // always 64 SPT, logically speaking
		int bsh;
		int bsm = 0;
		int exm = 0;
		int dsm = 0;
		int drm = 0;
		int alv0 = 0;
		int dxf;
		double siz = psize * ssz;
		siz /= 1024;
		siz /= 1024;

		for (bsh = minBsh; bsh <= 8; ++bsh) {
			bsm = (1 << bsh) - 1;
			dsm = cpmsecs >> bsh;
			if (dsm - 1 < 65536) {
				// TODO: could try for 0x100, but...
				if (dsm - 1 >= 256) {
					if (bsh < 4) {
						continue;
					}
					dxf = 2;
					exm = (1 << (bsh - 4)) - 1;
				} else {
					dxf = 1;
					exm = (1 << (bsh - 3)) - 1;
				}
				// now figure number of DIRENTs...
				drm = (dsm / (16 / dxf));
				drm = (int)Math.floor(drm * factor);
				// TODO pad some extra DIRENTs...
				int x = (drm + 3) / 4; // num of CP/M sectors required for dir
				x = (x + bsm) >> bsh; // num alloc blocks required for dir
				if (x > 16) {
					// can't represent this directory solution in ALV0
					continue;
				}
				alv0 = ~(0x0ffff >> x);
				drm = (x << bsh) * 4;
				break;
			}
		}
		// error check...
		if (bsh > 8) {
			System.err.format("No suitable BSH found for capacity %d\n", psize);
			return null;
		}
		System.out.format("Using DPB for partition %d = %gMB:\n" +
				"        %d\n" +
				"        %d,%d,%d\n" +
				"        %d-1, %d-1\n" +
				"        %02x, %02x\n" +
				"        %d,%d\n",
				cpmsecs, siz,
				64, bsh, bsm, exm, dsm, drm,
				(alv0 >> 8) & 0x0ff, alv0 & 0x0ff, 0, ff.resv_trks);
		dpb.spt = 64;
		dpb.bsh = bsh;
		dpb.blm = bsm;
		dpb.exm = exm;
		dpb.dsm = dsm - 1;
		dpb.drm = drm - 1;
		dpb.al0 = alv0;
		dpb.cks = 0; // fixed media, no CKS
		dpb.off = ff.resv_trks;
		dpb.poff = (start * ssz) / 128 + H89SASIFormatter.RESV_SECS;
		return dpb;
	}

	public static void main(String[] args) {
		new fmtsasi(args);
	}

	public fmtsasi(String[] args) {
		File fd;
		int c;
		int err = 0;
		int[] parts = new int[8];
		int nparts = 0;
		boolean overwrite = false;
		boolean nopart = false;
		boolean repair = false;
		boolean use_rest = false;
		boolean dump = false;
		float factor = 1.0f;
		int lat = 1; // not used?
		int bsh = 3;
		int disk = 0;
		int ssz = 256;
		CpmDpb[] dpbs = new CpmDpb[8];
		File[] preload = new File[8];
		int preloads = 0;

		if (args.length < 1) {
			usage(); // never returns
		}
		Arrays.fill(preload, null);
		for (c = 0; c < args.length; ++c) {
			if (args[c].equalsIgnoreCase("list")) {
				list(); // never returns
			} else if (args[c].equalsIgnoreCase("dump")) {
				dump = true;
				++c;
				break;
			} else if (args[c].equalsIgnoreCase("force")) {
				overwrite = true;
			} else if (args[c].equalsIgnoreCase("nopart")) {
				nopart = true;
			} else if (args[c].equalsIgnoreCase("repair")) {
				repair = true;
			} else if (args[c].startsWith("bsh=")) {
				bsh = Integer.decode(args[c].substring(4));
			} else if (args[c].startsWith("drm=")) {
				factor = Float.valueOf(args[c].substring(4));
				if (factor < 1.0) {
					factor = 1.0f;
				}
			} else if (args[c].startsWith("lat=")) {
				lat = Integer.decode(args[c].substring(4));
			} else if (args[c].startsWith("files=")) {
				int p = 0;
				File f;
				int comma = args[c].indexOf(',');
				if (comma < 0) {
					f = new File(args[c].substring(6));
				} else {
					try {
						p = Integer.valueOf(args[c].substring(6, comma));
					} catch (Exception ee) {}
					f = new File(args[c].substring(comma + 1));
				}
				if (!f.exists() || !f.isDirectory()) {
					System.err.format("No directory %s\n", f.getAbsolutePath());
				} else {
					preload[p] = f;
					++preloads;
				}
			} else if (args[c].startsWith("par=")) {
				if (nparts >= 8) {
					System.err.format("too many partitions\n");
					System.exit(1);
				}
				// if "0" then use rest of disk... must be last!
				if (use_rest) {
					System.err.format("Partition \"rest of disk\" is not last\n");
					System.exit(1);
				}
				int p = Integer.decode(args[c].substring(4));
				if (p == 0) {
					use_rest = true;
				}
				parts[nparts++] = p;
			} else {
				break; // done with options...
			}
		}
		// 'c' points to first non-option arg, i.e. image file name
		if (bsh < 3 || bsh > 8) {
			System.err.format("Invalid value for BSH: %d\n", bsh);
			System.exit(1);
		}
		if (c >= args.length) {
			usage(); // never returns
		}
		String name = args[c++];
		File fi = new File(name);
		if (dump) {
			dump(fi);
		}
		for (; c < args.length; ++c) {
			if (args[c].equalsIgnoreCase("XEBEC_ST506")) {
				disk = H89SASIFormatter.XEBEC_ST506;
			} else if (args[c].equalsIgnoreCase("XEBEC_ST412")) {
				disk = H89SASIFormatter.XEBEC_ST412;
			} else if (args[c].equalsIgnoreCase("XEBEC_CM5206")) {
				disk = H89SASIFormatter.XEBEC_CM5206;
			} else if (args[c].equalsIgnoreCase("XEBEC_CM5410")) {
				disk = H89SASIFormatter.XEBEC_CM5410;
			} else if (args[c].equalsIgnoreCase("XEBEC_CM5616")) {
				disk = H89SASIFormatter.XEBEC_CM5616;
			} else if (args[c].equalsIgnoreCase("XEBEC_RO201")) {
				disk = H89SASIFormatter.XEBEC_RO201;
			} else if (args[c].equalsIgnoreCase("XEBEC_RO202")) {
				disk = H89SASIFormatter.XEBEC_RO202;
			} else if (args[c].equalsIgnoreCase("XEBEC_RO203")) {
				disk = H89SASIFormatter.XEBEC_RO203;
			} else if (args[c].equalsIgnoreCase("XEBEC_RO204")) {
				disk = H89SASIFormatter.XEBEC_RO204;
			} else if (args[c].equalsIgnoreCase("ss")) {
				ssz = 512;
			} else {
				break; // done with format options...
			}
		}
		// 'c' points to first non-option arg
		if (c < args.length) {
			System.err.format("Invalid arguments at \"%\"\n", args[c]);
			usage(); // never returns
		}
		if (disk == 0) {
			System.err.format("Must specify a valid disk/controller combination\n");
			usage(); // never returns
		}
		H89SASIFormatter ff = new H89SASIFormatter(disk, ssz, lat, bsh, factor);
		if (ff.isError()) {
			System.err.format("Not a valid disk/controller config\n");
			usage(); // never returns
		}
		int avail = ff.getAvailable();
		if (nparts == 0) {
			// No partitions specified, make one big one...
			parts[nparts++] = avail;
		}
		int capacity = ff.getCapacity(); // bytes...
		double cap = capacity;
		cap /= 1024;
		cap /= 1024;
		System.err.format("Using disk %dx%d (%.1fMB) in %d partition(s)\n",
				capacity / ssz, ssz, cap, nparts);

		// parts[] are in native sectors (geom.ssz), but the CP/M partition
		// table is in CP/M sectors (128-byte sectors)!

		int t = 0;
		for (int x = 0; x < nparts; ++x) {
			if (parts[x] == 0) {
				parts[x] = avail - t;
			}
			dpbs[x] = setcpmdpb(ff, t, parts[x], ssz, bsh, factor);
			t += parts[x];
			if (t > avail) {
				System.err.format("Partition %d exceeds available capacity %d\n", x, avail);
				System.exit(1);
			}
		}
		if (t < avail) {
			System.err.format("Unused sectors: %d\n", avail - t);
		}

		if (fi.exists() && !overwrite) {
			System.err.format("File %s already exists\n", name);
			System.exit(1);
		}
		if (!fi.exists() && repair) {
			System.err.format("File to repair %s does not exists\n", name);
			System.exit(1);
		}

		// TODO: how to ensure APPEND mode?
		File ddfi = new File("diskdefs");

		if (!nopart && !ff.setPartitions(parts, nparts, ddfi, dpbs)) {
			System.exit(1);
		}
		if (!ff.format(fi, repair)) {
			System.exit(1);
		}
		System.out.format("Formatted %s successfully\n", name);
		if (preloads > 0) {
			CpmCreateFile cf = new CpmCreateFile(fi);
			if (cf.isError()) {
				System.err.format("Skipping preloads due to CpmCreateFile error\n");
				System.exit(1);
			}
			for (int x = 0; x < nparts; ++x) {
				if (preload[x] != null) {
					// preload[x] is a directory...
					cf.doPreload(preload[x], dpbs[x]);
				}
			}
			cf.close();
		}
	}
}
