#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

// Program to examine disk images from format.jar or fmtsasi.jar and
// generate a diskdef for use with cpmtools. Uses the MMS-style
// partition table for fmtsasi.jar images.
//
// NOTE: Kaypro uses the winchester disk as two partitions
// with every other pair of tracks belonging to a given partition.
// This cannot be represented to cpmtools (AFAIK).

// Command syntax: getdiskdef <image-file> [<partition-num>]

int mediaSize_m;
int secSize_m;
int numSectors_m;
int numSides_m;
int numTracks_m;
int doubleDensity_m;
int interlaced_m;
int mediaLat_m;
int mediaCyl;
int mediaHead;
int mediaSsz;
int mediaSpt;
int mediaSec_m = 0;
int result;

int checkHeader(char *buf, int n) {
    char *b = buf;
    result = 0;

    while (*b != '\n' && *b != '\0' && b - buf < n) {
        char *e;
        int p = strtoul(b, &e, 0);

        switch (tolower(*e)) {
        case 'm':
            if (p != 5 && p != 8) {
                break;
            }
            result |= 0x01;
            mediaSize_m = p;
            break;

        case 'z':
            if (p != 128 && p != 256 && p != 512 && p != 1024) {
                break;
            }
            result |= 0x02;
            secSize_m = p;
            mediaSsz = p;
            break;

        case 'p':
            if (p == 0) {
                break;
            }
            result |= 0x04;
            numSectors_m = p;
            mediaSpt = p;
            break;

        case 's':
            if (p < 1 || p > 2) {
                break;
            }
            result |= 0x08;
            numSides_m = p;
            break;

        case 't':
            if (p == 0) {
                break;
            }
            result |= 0x10;
            numTracks_m = p;
            break;

        case 'd':
            result |= 0x20;
            doubleDensity_m = p;
            break;

        case 'i':
            result |= 0x40;
            interlaced_m = p;
            break;

        case 'c':
            if (p == 0) {
                break;
            }
            result |= 0x80;
            mediaCyl = p;
            break;

        case 'h': /* might be SASI or Z17... */
	    if (result & 0x01) {
            	mediaSec_m = p;
                break;
	    }
            if (p == 0) {
                break;
            }
            result |= 0x0100;
            mediaHead = p;
            break;

        case 'l': // optional ?
            mediaLat_m = p;
            break;

        default:
            return 0;
        }

        b = e + 1;
    }

    return (result == 0x7f || result == 0x186);
}

struct dpb {
	unsigned short spt;
	unsigned char bsh, bsm, exm;
	unsigned short dsm, drm;
	unsigned char alv0, alv1;
	unsigned short cks;
	unsigned short restrk;
	unsigned char pad[6]; // non-CP/M stuff
} __attribute__ ((packed));

struct sasi_fdisk {
	unsigned char jmp[3];
	unsigned char drive;
	unsigned char ctrl;
	unsigned char dcharistics[8];
	unsigned char assdtype[6];
	unsigned char nparts;
	unsigned char parts[9][3];
	struct dpb dpbs[8];
} __attribute__ ((packed));


float check_os(FILE *fp, struct dpb *dpb, int offset) {
	float os = 2.2;
	unsigned char buf[128];
	long off = offset * 128;
	off += dpb->restrk * dpb->spt * 128;
	fseek(fp, off, SEEK_SET);
	size_t rc = fread(buf, sizeof(buf), 1, fp);
	if (rc == 1 && buf[96] == 0x21) {
		os = 3;
	}
	return os;
}

void get_sasi_diskdef(FILE *fp, void *ptab, int partno) {
	struct sasi_fdisk *fdisk = (struct sasi_fdisk *)ptab;
if (0) {
int x;
for (x = 0; x < fdisk->nparts; ++x) {
	int y = (fdisk->parts[x][0] << 16) |
			(fdisk->parts[x][1] << 8) |
			fdisk->parts[x][2];
printf("%d: %d\n", x, y);
}
for (x = 0; x < fdisk->nparts; ++x) {
printf("%d dpb: %d %d %d %d %d %d %02x%02x %d %d\n", x,
fdisk->dpbs[x].spt,
fdisk->dpbs[x].bsh, fdisk->dpbs[x].bsm, fdisk->dpbs[x].exm,
fdisk->dpbs[x].dsm, fdisk->dpbs[x].drm,
fdisk->dpbs[x].alv0, fdisk->dpbs[x].alv1,
fdisk->dpbs[x].cks,
fdisk->dpbs[x].restrk);
}
}
	if (partno >= fdisk->nparts) {
		return;
	}
	int offset = (fdisk->parts[partno][0] << 16) |
			(fdisk->parts[partno][1] << 8) |
			fdisk->parts[partno][2];
	int nextoff = (fdisk->parts[partno + 1][0] << 16) |
			(fdisk->parts[partno + 1][1] << 8) |
			fdisk->parts[partno + 1][2];
	int sxf = mediaSsz / 128;
	if (nextoff == 0) {
		nextoff = (mediaCyl * mediaHead * mediaSpt) * sxf;
	}
	float os = check_os(fp, &fdisk->dpbs[partno], offset);
	printf("diskdef\n"
		"  seclen %d\n  tracks %d\n  sectrk %d\n"
		"  blocksize %d\n  maxdir %d\n  skew 0\n"
		"  boottrk %d\n  offset %d\n  os %g\nend\n",
		mediaSsz, (nextoff - offset) / fdisk->dpbs[partno].spt,
		fdisk->dpbs[partno].spt / sxf,
		(1 << fdisk->dpbs[partno].bsh) * 128,
		fdisk->dpbs[partno].drm + 1,
		fdisk->dpbs[partno].restrk,
		offset * 128, os);
}

enum fmts { MMS, Z17, M47, Z37, Z47, Z67, Z37X, Z47X, KAYPRO, HIDENS, FMTMAX };
enum medias { F525, F8, MEDMAX };
enum densities { SD, DD, QD, DENSMAX };
enum sides { SS, DS, SIDSMAX };
enum tpis { ST, DT, QT, TPISMAX };

// pad[0] is software sector skew (latency) for disk format.
// This is used to eliminate skew from formats that show
// "hard skew" i.e. the skew used during formatting.
// TODO: need to fix inconsistency - it may have just been
// early versions of H89FloppyFormatter that did that.
struct dpb floppy_dpbs[MEDMAX][FMTMAX][DENSMAX][SIDSMAX][TPISMAX] = {
[F525][Z17][SD][SS][ST] = { 20, 3,  7, 0,  91,  63, 0xc0, 0x00, 16, 3,
			{4, 0, 0, 0, 0, 0}},
[F525][Z17][SD][DS][ST] = { 20, 3,  7, 0, 181,  63, 0xc0, 0x00, 16, 3,
			{4, 0, 0, 0, 0, 0}},
[F525][Z17][SD][SS][DT] = { 20, 4, 15, 1,  95, 127, 0xc0, 0x00, 32, 3,
			{4, 0, 0, 0, 0, 0}},
[F525][Z17][SD][DS][DT] = { 20, 4, 15, 1, 185, 127, 0xc0, 0x00, 32, 3,
			{4, 0, 0, 0, 0, 0}},

[F8][M47][DD][SS][ST] = { 64, 4, 15, 0, 299, 191, 0xe0, 0x00, 48, 2,
			{4, 0, 0, 0, 0, 0}},
[F8][M47][DD][DS][ST] = { 64, 4, 15, 0, 607, 191, 0xe0, 0x00, 48, 2,
			{4, 0, 0, 0, 0, 0}},

[F525][MMS][DD][SS][ST] = { 36, 4, 15, 1,  82,  95, 0xc0, 0x00, 24, 3,
			{0, 0, 0, 0, 0, 0}},
[F525][MMS][DD][DS][ST] = { 36, 4, 15, 1, 172,  95, 0xc0, 0x00, 24, 3,
			{0, 0, 0, 0, 0, 0}},
[F525][MMS][DD][SS][DT] = { 36, 5, 31, 3,  85, 127, 0x80, 0x00, 32, 3,
			{0, 0, 0, 0, 0, 0}},
[F525][MMS][DD][DS][DT] = { 36, 5, 31, 3, 175, 127, 0x80, 0x00, 32, 3,
			{0, 0, 0, 0, 0, 0}},
[F8][MMS][DD][SS][ST]   = { 64, 4, 15, 0, 299, 191, 0xe0, 0x00, 48, 2,
			{0, 0, 0, 0, 0, 0}},
[F8][MMS][DD][DS][ST]   = { 64, 4, 15, 0, 607, 191, 0xe0, 0x00, 48, 2,
			{0, 0, 0, 0, 0, 0}},

[F525][Z37][SD][SS][ST] = { 20, 3,  7, 0,  91,  63, 0xc0, 0x00, 16, 3,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37][SD][DS][ST] = { 20, 3,  7, 0, 187, 127, 0xf0, 0x00, 32, 3,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37][SD][SS][DT] = { 20, 3,  7, 0, 191,  63, 0xc0, 0x00, 16, 3,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37][SD][DS][DT] = { 20, 4, 15, 1, 195, 127, 0xc0, 0x00, 32, 3,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37][DD][SS][ST] = { 32, 4, 15, 1, 151, 127, 0xf0, 0x00, 32, 2,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37][DD][DS][ST] = { 32, 4, 15, 0, 155, 255, 0xf0, 0x00, 64, 2,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37][DD][SS][DT] = { 32, 4, 15, 1, 155, 127, 0xc0, 0x00, 32, 2,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37][DD][DS][DT] = { 32, 4, 15, 0, 315, 255, 0xf0, 0x00, 64, 2,
			{0, 0, 0, 0, 0, 0}},

[F525][Z37X][DD][SS][ST] = { 40, 3,  7, 0, 185, 127, 0xf0, 0x00, 32, 2,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37X][DD][DS][ST] = { 40, 4, 15, 0, 194, 255, 0xf0, 0x00, 64, 2,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37X][DD][SS][DT] = { 40, 4, 15, 1, 194, 127, 0xc0, 0x00, 32, 2,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37X][DD][DS][DT] = { 40, 4, 15, 0, 394, 255, 0xf0, 0x00, 64, 2,
			{0, 0, 0, 0, 0, 0}},

[F8][Z47][DD][SS][ST]    = { 52, 4, 15, 0, 242, 127, 0xc0, 0x00, 32, 2,
			{9, 0, 0, 0, 0, 0}},
[F8][Z47][DD][DS][ST]    = { 52, 4, 15, 0, 493, 255, 0xf0, 0x00, 64, 2,
			{9, 0, 0, 0, 0, 0}},
[F8][Z47X][DD][SS][ST]   = { 64, 4, 15, 0, 299, 127, 0xc0, 0x00, 32, 2,
			{4, 0, 0, 0, 0, 0}},
[F8][Z47X][DD][DS][ST]   = { 64, 4, 15, 0, 607, 255, 0xf0, 0x00, 64, 2,
			{4, 0, 0, 0, 0, 0}},

[F525][KAYPRO][DD][SS][ST] = { 40, 3,  7, 0, 194, 63, 0xf0, 0x00, 16, 1,
			{0, 0, 0, 0, 0, 0}},
[F525][KAYPRO][DD][DS][ST] = { 40, 4, 15, 1, 196, 63, 0xc0, 0x00, 16, 1,
			{0, 0, 0, 0, 0, 0}},

[F525][HIDENS][QD][DS][QT] = { 68, 4, 15, 0,1351,1023,0xff, 0xff,256, 2,
			{5, 0, 0, 0, 0, 0}},
};

void get_floppy_diskdef(FILE *fp) {
	enum fmts fmt;
	enum medias media = mediaSize_m == 5 ? F525 : F8;
	enum densities dens = SD;
	if (doubleDensity_m > 1) {
		dens = QD;
	} else if (doubleDensity_m > 0) {
		dens = DD;
	}
	enum sides sid = numSides_m > 1 ? DS : SS;
	enum tpis tpi = ST;
	if (media == F525) {
		if (numTracks_m > 80) {
			tpi = QT;
		} else if (numTracks_m > 40) {
			tpi = DT;
		}
	}
	if (media == F8) {
		fmt = Z47;
	} else {
		fmt = Z37;
	}
	if (doubleDensity_m > 0 && secSize_m == 512 && numSectors_m == 9) {
		fmt = MMS;
	} else if (doubleDensity_m > 1  && secSize_m == 512 && numSectors_m == 17) {
		fmt = HIDENS;
	}
	if (doubleDensity_m && secSize_m == 512 &&
			numSectors_m == 10 && interlaced_m == 2) {
		fmt = KAYPRO;
	}
	// can't tell M47 from Z47X... just use Z47X...
	// except by interlaced?
	if (media == F8 && secSize_m == 1024 && numSectors_m == 16) {
		fmt = interlaced_m ? Z47X : M47;
	}
	if (secSize_m == 1024) {
		fmt = media == F8 ? Z47X : Z37X;
	}
	if (mediaSec_m == numSectors_m) {
		fmt = Z17;
	}
	struct dpb *dpb = &floppy_dpbs[media][fmt][dens][sid][tpi];
	if (dpb->pad[0] != mediaLat_m) {
		mediaLat_m = 0;
	}

	float os = check_os(fp, dpb, 0L);
	printf("diskdef\n"
		"  seclen %d\n  tracks %d\n  sectrk %d\n"
		"  blocksize %d\n  maxdir %d\n  skew %d\n"
		"  boottrk %d\n  os %g\nend\n",
		mediaSsz, numTracks_m * numSides_m,
		dpb->spt / (secSize_m / 128),
		(1 << dpb->bsh) * 128,
		dpb->drm + 1, mediaLat_m,
		dpb->restrk, os);
}

int main(int argc, char **argv) {
	char buf[128];
	int partno = 0;

	if (argc < 2 || argc > 3) {
		fprintf(stderr, "Syntax: %s <image-file> [<partition-num>]\n", argv[0]);
		exit(1);
	}
	if (argc > 2) {
		partno = strtoul(argv[2], NULL, 0);
	}

	FILE *fp = fopen(argv[1], "r");
	if (fp == NULL) {
		perror(argv[1]);
		exit(1);
	}
	fseek(fp, 0L, SEEK_END);
	long end = ftell(fp);
	fseek(fp, end - sizeof(buf), SEEK_SET);
	size_t rc = fread(buf, sizeof(buf), 1, fp);
	if (rc != 1 || !checkHeader(buf, sizeof(buf))) {
		// silently fail if not recognized?
		fprintf(stderr, "Failed to recognize image %ld %04x\n", rc, result);
		fclose(fp);
		exit(1);
	}
	if (result == 0x7f) {
		// Floppy diskette...
		get_floppy_diskdef(fp);
	} else {
		// SASI disk, possible partition table.
		rewind(fp);
		char ptab[512];
		fread(ptab, sizeof(ptab), 1, fp);
		get_sasi_diskdef(fp, ptab, partno);
	}
	fclose(fp);
	return 0;
}
